@@ -1,3 +1,42 @@
+/*
+ * libeio implementation
+ *
+ * Copyright (c) 2007,2008 Marc Alexander Lehmann <libeio@schmorp.de>
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without modifica-
+ * tion, are permitted provided that the following conditions are met:
+ * 
+ *   1.  Redistributions of source code must retain the above copyright notice,
+ *       this list of conditions and the following disclaimer.
+ * 
+ *   2.  Redistributions in binary form must reproduce the above copyright
+ *       notice, this list of conditions and the following disclaimer in the
+ *       documentation and/or other materials provided with the distribution.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
+ * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MER-
+ * CHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
+ * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPE-
+ * CIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
+ * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
+ * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTH-
+ * ERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
+ * OF THE POSSIBILITY OF SUCH DAMAGE.
+ *
+ * Alternatively, the contents of this file may be used under the terms of
+ * the GNU General Public License ("GPL") version 2 or any later version,
+ * in which case the provisions of the GPL are applicable instead of
+ * the above. If you wish to allow the use of your version of this file
+ * only under the terms of the GPL and not to allow others to use your
+ * version of this file under the BSD license, indicate your decision
+ * by deleting the provisions above and replace them with the notice
+ * and other provisions required by the GPL. If you do not delete the
+ * provisions above, a recipient may use your version of this file under
+ * either the BSD or the GPL.
+ */
+
 #include "eio.h"
 #include "xthread.h"
 
@@ -72,20 +111,48 @@
 
 #define dBUF	 				\
   char *eio_buf;				\
-  X_LOCK (wrklock);				\
+  ETP_WORKER_LOCK (self);			\
   self->dbuf = eio_buf = malloc (EIO_BUFSIZE);	\
-  X_UNLOCK (wrklock);				\
+  ETP_WORKER_UNLOCK (self);			\
   errno = ENOMEM;				\
   if (!eio_buf)					\
     return -1;
 
 #define EIO_TICKS ((1000000 + 1023) >> 10)
 
-static void (*want_poll_cb) (void);
-static void (*done_poll_cb) (void);
+/*****************************************************************************/
 
-static unsigned int max_poll_time = 0;
-static unsigned int max_poll_reqs = 0;
+#define ETP_PRI_MIN EIO_PRI_MIN
+#define ETP_PRI_MAX EIO_PRI_MAX
+
+struct etp_worker;
+
+#define ETP_REQ eio_req
+#define ETP_DESTROY(req) eio_destroy (req)
+static int eio_finish (eio_req *req);
+#define ETP_FINISH(req)  eio_finish (req)
+static void eio_execute (struct etp_worker *self, eio_req *req);
+#define ETP_EXECUTE(wrk,req) eio_execute (wrk,req)
+
+#define ETP_WORKER_CLEAR(req)	\
+  if (wrk->dbuf)		\
+    {				\
+      free (wrk->dbuf);		\
+      wrk->dbuf = 0;		\
+    }				\
+				\
+  if (wrk->dirp)		\
+    {				\
+      closedir (wrk->dirp);	\
+      wrk->dirp = 0;		\
+    }
+#define ETP_WORKER_COMMON \
+  void *dbuf;	\
+  DIR *dirp;
+
+/*****************************************************************************/
+
+#define ETP_NUM_PRI (ETP_PRI_MAX - ETP_PRI_MIN + 1)
 
 /* calculcate time difference in ~1/EIO_TICKS of a second */
 static int tvdiff (struct timeval *tv1, struct timeval *tv2)
@@ -96,40 +163,48 @@ static int tvdiff (struct timeval *tv1, struct timeval *tv2)
 
 static unsigned int started, idle, wanted = 4;
 
-/* worker threads management */
-static mutex_t wrklock = X_MUTEX_INIT;
+static void (*want_poll_cb) (void);
+static void (*done_poll_cb) (void);
+ 
+static unsigned int max_poll_time;     /* reslock */
+static unsigned int max_poll_reqs;     /* reslock */
 
-typedef struct worker
+static volatile unsigned int nreqs;    /* reqlock */
+static volatile unsigned int nready;   /* reqlock */
+static volatile unsigned int npending; /* reqlock */
+static volatile unsigned int max_idle = 4;
+
+static mutex_t wrklock = X_MUTEX_INIT;
+static mutex_t reslock = X_MUTEX_INIT;
+static mutex_t reqlock = X_MUTEX_INIT;
+static cond_t  reqwait = X_COND_INIT;
+
+typedef struct etp_worker
 {
   /* locked by wrklock */
-  struct worker *prev, *next;
+  struct etp_worker *prev, *next;
 
   thread_t tid;
 
   /* locked by reslock, reqlock or wrklock */
-  eio_req *req; /* currently processed request */
-  void *dbuf;
-  DIR *dirp;
-} worker;
+  ETP_REQ *req; /* currently processed request */
 
-static worker wrk_first = { &wrk_first, &wrk_first, 0 };
+  ETP_WORKER_COMMON
+} etp_worker;
 
-static void worker_clear (worker *wrk)
+static etp_worker wrk_first = { &wrk_first, &wrk_first, 0 }; /* NOT etp */
+
+#define ETP_WORKER_LOCK(wrk)   X_LOCK   (wrklock)
+#define ETP_WORKER_UNLOCK(wrk) X_UNLOCK (wrklock)
+
+/* worker threads management */
+
+static void etp_worker_clear (etp_worker *wrk)
 {
-  if (wrk->dirp)
-    {
-      closedir (wrk->dirp);
-      wrk->dirp = 0;
-    }
-
-  if (wrk->dbuf)
-    {
-      free (wrk->dbuf);
-      wrk->dbuf = 0;
-    }
+  ETP_WORKER_CLEAR (wrk);
 }
 
-static void worker_free (worker *wrk)
+static void etp_worker_free (etp_worker *wrk)
 {
   wrk->next->prev = wrk->prev;
   wrk->prev->next = wrk->next;
@@ -137,19 +212,16 @@ static void worker_free (worker *wrk)
   free (wrk);
 }
 
-static volatile unsigned int nreqs, nready, npending;
-static volatile unsigned int max_idle = 4;
-
-static mutex_t reslock = X_MUTEX_INIT;
-static mutex_t reqlock = X_MUTEX_INIT;
-static cond_t  reqwait = X_COND_INIT;
-
-unsigned int eio_nreqs (void)
+static unsigned int etp_nreqs (void)
 {
-  return nreqs;
+  int retval;
+  if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
+  retval = nreqs;
+  if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
+  return retval;
 }
 
-unsigned int eio_nready (void)
+static unsigned int etp_nready (void)
 {
   unsigned int retval;
 
@@ -160,7 +232,7 @@ unsigned int eio_nready (void)
   return retval;
 }
 
-unsigned int eio_npending (void)
+static unsigned int etp_npending (void)
 {
   unsigned int retval;
 
@@ -171,7 +243,7 @@ unsigned int eio_npending (void)
   return retval;
 }
 
-unsigned int eio_nthreads (void)
+static unsigned int etp_nthreads (void)
 {
   unsigned int retval;
 
@@ -188,14 +260,14 @@ unsigned int eio_nthreads (void)
  * per shift, the most expensive operation.
  */
 typedef struct {
-  eio_req *qs[EIO_NUM_PRI], *qe[EIO_NUM_PRI]; /* qstart, qend */
+  ETP_REQ *qs[ETP_NUM_PRI], *qe[ETP_NUM_PRI]; /* qstart, qend */
   int size;
-} reqq;
+} etp_reqq;
 
-static reqq req_queue;
-static reqq res_queue;
+static etp_reqq req_queue;
+static etp_reqq res_queue;
 
-static int reqq_push (reqq *q, eio_req *req)
+static int reqq_push (etp_reqq *q, ETP_REQ *req)
 {
   int pri = req->pri;
   req->next = 0;
@@ -211,7 +283,7 @@ static int reqq_push (reqq *q, eio_req *req)
   return q->size++;
 }
 
-static eio_req *reqq_shift (reqq *q)
+static ETP_REQ *reqq_shift (etp_reqq *q)
 {
   int pri;
 
@@ -220,7 +292,7 @@ static eio_req *reqq_shift (reqq *q)
 
   --q->size;
 
-  for (pri = EIO_NUM_PRI; pri--; )
+  for (pri = ETP_NUM_PRI; pri--; )
     {
       eio_req *req = q->qs[pri];
 
@@ -236,6 +308,265 @@ static eio_req *reqq_shift (reqq *q)
   abort ();
 }
 
+static void etp_atfork_prepare (void)
+{
+  X_LOCK (wrklock);
+  X_LOCK (reqlock);
+  X_LOCK (reslock);
+#if !HAVE_PREADWRITE
+  X_LOCK (preadwritelock);
+#endif
+#if !HAVE_READDIR_R
+  X_LOCK (readdirlock);
+#endif
+}
+
+static void etp_atfork_parent (void)
+{
+#if !HAVE_READDIR_R
+  X_UNLOCK (readdirlock);
+#endif
+#if !HAVE_PREADWRITE
+  X_UNLOCK (preadwritelock);
+#endif
+  X_UNLOCK (reslock);
+  X_UNLOCK (reqlock);
+  X_UNLOCK (wrklock);
+}
+
+static void etp_atfork_child (void)
+{
+  ETP_REQ *prv;
+
+  while (prv = reqq_shift (&req_queue))
+    ETP_DESTROY (prv);
+
+  while (prv = reqq_shift (&res_queue))
+    ETP_DESTROY (prv);
+
+  while (wrk_first.next != &wrk_first)
+    {
+      etp_worker *wrk = wrk_first.next;
+
+      if (wrk->req)
+        ETP_DESTROY (wrk->req);
+
+      etp_worker_clear (wrk);
+      etp_worker_free (wrk);
+    }
+
+  started  = 0;
+  idle     = 0;
+  nreqs    = 0;
+  nready   = 0;
+  npending = 0;
+
+  etp_atfork_parent ();
+}
+
+static void
+etp_once_init (void)
+{    
+  X_THREAD_ATFORK (etp_atfork_prepare, etp_atfork_parent, etp_atfork_child);
+}
+
+static int
+etp_init (void (*want_poll)(void), void (*done_poll)(void))
+{
+  static pthread_once_t doinit = PTHREAD_ONCE_INIT;
+
+  pthread_once (&doinit, etp_once_init);
+
+  want_poll_cb = want_poll;
+  done_poll_cb = done_poll;
+}
+
+X_THREAD_PROC (etp_proc);
+
+static void etp_start_thread (void)
+{
+  etp_worker *wrk = calloc (1, sizeof (etp_worker));
+
+  /*TODO*/
+  assert (("unable to allocate worker thread data", wrk));
+
+  X_LOCK (wrklock);
+
+  if (thread_create (&wrk->tid, etp_proc, (void *)wrk))
+    {
+      wrk->prev = &wrk_first;
+      wrk->next = wrk_first.next;
+      wrk_first.next->prev = wrk;
+      wrk_first.next = wrk;
+      ++started;
+    }
+  else
+    free (wrk);
+
+  X_UNLOCK (wrklock);
+}
+
+static void etp_maybe_start_thread (void)
+{
+  if (etp_nthreads () >= wanted)
+    return;
+  
+  /* todo: maybe use idle here, but might be less exact */
+  if (0 <= (int)etp_nthreads () + (int)etp_npending () - (int)etp_nreqs ())
+    return;
+
+  etp_start_thread ();
+}
+
+static void etp_end_thread (void)
+{
+  eio_req *req = calloc (1, sizeof (eio_req));
+
+  req->type = -1;
+  req->pri  = ETP_PRI_MAX - ETP_PRI_MIN;
+
+  X_LOCK (reqlock);
+  reqq_push (&req_queue, req);
+  X_COND_SIGNAL (reqwait);
+  X_UNLOCK (reqlock);
+
+  X_LOCK (wrklock);
+  --started;
+  X_UNLOCK (wrklock);
+}
+
+static int etp_poll (void)
+{
+  unsigned int maxreqs;
+  unsigned int maxtime;
+  struct timeval tv_start, tv_now;
+
+  X_LOCK (reslock);
+  maxreqs = max_poll_reqs;
+  maxtime = max_poll_time;
+  X_UNLOCK (reslock);
+
+  if (maxtime)
+    gettimeofday (&tv_start, 0);
+
+  for (;;)
+    {
+      ETP_REQ *req;
+
+      etp_maybe_start_thread ();
+
+      X_LOCK (reslock);
+      req = reqq_shift (&res_queue);
+
+      if (req)
+        {
+          --npending;
+
+          if (!res_queue.size && done_poll_cb)
+            done_poll_cb ();
+        }
+
+      X_UNLOCK (reslock);
+
+      if (!req)
+        return 0;
+
+      X_LOCK (reqlock);
+      --nreqs;
+      X_UNLOCK (reqlock);
+
+      if (req->type == EIO_GROUP && req->size)
+        {
+          req->int1 = 1; /* mark request as delayed */
+          continue;
+        }
+      else
+        {
+          int res = ETP_FINISH (req);
+          if (res)
+            return res;
+        }
+
+      if (maxreqs && !--maxreqs)
+        break;
+
+      if (maxtime)
+        {
+          gettimeofday (&tv_now, 0);
+
+          if (tvdiff (&tv_start, &tv_now) >= maxtime)
+            break;
+        }
+    }
+
+  errno = EAGAIN;
+  return -1;
+}
+
+static void etp_cancel (ETP_REQ *req)
+{
+  X_LOCK   (wrklock);
+  req->flags |= EIO_FLAG_CANCELLED;
+  X_UNLOCK (wrklock);
+
+  eio_grp_cancel (req);
+}
+
+static void etp_submit (ETP_REQ *req)
+{
+  req->pri -= ETP_PRI_MIN;
+
+  if (req->pri < ETP_PRI_MIN - ETP_PRI_MIN) req->pri = ETP_PRI_MIN - ETP_PRI_MIN;
+  if (req->pri > ETP_PRI_MAX - ETP_PRI_MIN) req->pri = ETP_PRI_MAX - ETP_PRI_MIN;
+
+  X_LOCK (reqlock);
+  ++nreqs;
+  ++nready;
+  reqq_push (&req_queue, req);
+  X_COND_SIGNAL (reqwait);
+  X_UNLOCK (reqlock);
+
+  etp_maybe_start_thread ();
+}
+
+static void etp_set_max_poll_time (double nseconds)
+{
+  if (WORDACCESS_UNSAFE) X_LOCK   (reslock);
+  max_poll_time = nseconds;
+  if (WORDACCESS_UNSAFE) X_UNLOCK (reslock);
+}
+
+static void etp_set_max_poll_reqs (unsigned int maxreqs)
+{
+  if (WORDACCESS_UNSAFE) X_LOCK   (reslock);
+  max_poll_reqs = maxreqs;
+  if (WORDACCESS_UNSAFE) X_UNLOCK (reslock);
+}
+
+static void etp_set_max_idle (unsigned int nthreads)
+{
+  if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
+  max_idle = nthreads <= 0 ? 1 : nthreads;
+  if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
+}
+
+static void etp_set_min_parallel (unsigned int nthreads)
+{
+  if (wanted < nthreads)
+    wanted = nthreads;
+}
+
+static void etp_set_max_parallel (unsigned int nthreads)
+{
+  if (wanted > nthreads)
+    wanted = nthreads;
+
+  while (started > wanted)
+    etp_end_thread ();
+}
+
+/*****************************************************************************/
+
 static void grp_try_feed (eio_req *grp)
 {
   while (grp->size < grp->int2 && !EIO_CANCELLED (grp))
@@ -253,8 +584,6 @@ static void grp_try_feed (eio_req *grp)
     }
 }
 
-static int eio_finish (eio_req *req);
-
 static int grp_dec (eio_req *grp)
 {
   --grp->size;
@@ -312,173 +641,62 @@ void eio_grp_cancel (eio_req *grp)
 
 void eio_cancel (eio_req *req)
 {
-  X_LOCK   (wrklock);
-  req->flags |= EIO_FLAG_CANCELLED;
-  X_UNLOCK (wrklock);
-
-  eio_grp_cancel (req);
-}
-
-X_THREAD_PROC (eio_proc);
-
-static void start_thread (void)
-{
-  worker *wrk = calloc (1, sizeof (worker));
-
-  /*TODO*/
-  assert (("unable to allocate worker thread data", wrk));
-
-  X_LOCK (wrklock);
-
-  if (thread_create (&wrk->tid, eio_proc, (void *)wrk))
-    {
-      wrk->prev = &wrk_first;
-      wrk->next = wrk_first.next;
-      wrk_first.next->prev = wrk;
-      wrk_first.next = wrk;
-      ++started;
-    }
-  else
-    free (wrk);
-
-  X_UNLOCK (wrklock);
-}
-
-static void maybe_start_thread (void)
-{
-  if (eio_nthreads () >= wanted)
-    return;
-  
-  /* todo: maybe use idle here, but might be less exact */
-  if (0 <= (int)eio_nthreads () + (int)eio_npending () - (int)eio_nreqs ())
-    return;
-
-  start_thread ();
+  etp_cancel (req);
 }
 
 void eio_submit (eio_req *req)
 {
-  ++nreqs;
-
-  X_LOCK (reqlock);
-  ++nready;
-  reqq_push (&req_queue, req);
-  X_COND_SIGNAL (reqwait);
-  X_UNLOCK (reqlock);
-
-  maybe_start_thread ();
+  etp_submit (req);
 }
 
-static void end_thread (void)
+unsigned int eio_nreqs (void)
 {
-  eio_req *req = calloc (1, sizeof (eio_req));
+  return etp_nreqs ();
+}
 
-  req->type = EIO_QUIT;
-  req->pri  = EIO_PRI_MAX + EIO_PRI_BIAS;
+unsigned int eio_nready (void)
+{
+  return etp_nready ();
+}
 
-  X_LOCK (reqlock);
-  reqq_push (&req_queue, req);
-  X_COND_SIGNAL (reqwait);
-  X_UNLOCK (reqlock);
+unsigned int eio_npending (void)
+{
+  return etp_npending ();
+}
 
-  X_LOCK (wrklock);
-  --started;
-  X_UNLOCK (wrklock);
+unsigned int eio_nthreads (void)
+{
+  return etp_nthreads ();
 }
 
 void eio_set_max_poll_time (double nseconds)
 {
-  if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
-  max_poll_time = nseconds;
-  if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
+  etp_set_max_poll_time (nseconds);
 }
 
 void eio_set_max_poll_reqs (unsigned int maxreqs)
 {
-  if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
-  max_poll_reqs = maxreqs;
-  if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
+  etp_set_max_poll_reqs (maxreqs);
 }
 
 void eio_set_max_idle (unsigned int nthreads)
 {
-  if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
-  max_idle = nthreads <= 0 ? 1 : nthreads;
-  if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
+  etp_set_max_idle (nthreads);
 }
 
 void eio_set_min_parallel (unsigned int nthreads)
 {
-  if (wanted < nthreads)
-    wanted = nthreads;
+  etp_set_min_parallel (nthreads);
 }
 
 void eio_set_max_parallel (unsigned int nthreads)
 {
-  if (wanted > nthreads)
-    wanted = nthreads;
-
-  while (started > wanted)
-    end_thread ();
+  etp_set_max_parallel (nthreads);
 }
 
 int eio_poll (void)
 {
-  int maxreqs = max_poll_reqs;
-  struct timeval tv_start, tv_now;
-  eio_req *req;
-
-  if (max_poll_time)
-    gettimeofday (&tv_start, 0);
-
-  for (;;)
-    {
-      maybe_start_thread ();
-
-      X_LOCK (reslock);
-      req = reqq_shift (&res_queue);
-
-      if (req)
-        {
-          --npending;
-
-          if (!res_queue.size && done_poll_cb)
-            done_poll_cb ();
-        }
-
-      X_UNLOCK (reslock);
-
-      if (!req)
-        return 0;
-
-      --nreqs;
-
-      if (req->type == EIO_GROUP && req->size)
-        {
-          req->int1 = 1; /* mark request as delayed */
-          continue;
-        }
-      else
-        {
-          int res = eio_finish (req);
-          if (res)
-            return res;
-        }
-
-      if (maxreqs && !--maxreqs)
-        break;
-
-      if (max_poll_time)
-        {
-          gettimeofday (&tv_now, 0);
-
-          if (tvdiff (&tv_start, &tv_now) >= max_poll_time)
-            break;
-        }
-    }
-
-  errno = EAGAIN;
-  return -1;
+  return etp_poll ();
 }
 
 /*****************************************************************************/
@@ -618,7 +836,7 @@ eio__readdir_r (DIR *dirp, EIO_STRUCT_DIRENT *ent, EIO_STRUCT_DIRENT **res)
 
 /* sendfile always needs emulation */
 static ssize_t
-eio__sendfile (int ofd, int ifd, off_t offset, size_t count, worker *self)
+eio__sendfile (int ofd, int ifd, off_t offset, size_t count, etp_worker *self)
 {
   ssize_t res;
 
@@ -712,7 +930,7 @@ eio__sendfile (int ofd, int ifd, off_t offset, size_t count, worker *self)
 
 /* read a full directory */
 static void
-eio__scandir (eio_req *req, worker *self)
+eio__scandir (eio_req *req, etp_worker *self)
 {
   DIR *dirp;
   union
@@ -789,19 +1007,17 @@ eio__scandir (eio_req *req, worker *self)
         }					\
     }
 
-X_THREAD_PROC (eio_proc)
+X_THREAD_PROC (etp_proc)
 {
-  eio_req *req;
+  ETP_REQ *req;
   struct timespec ts;
-  worker *self = (worker *)thr_arg;
+  etp_worker *self = (etp_worker *)thr_arg;
 
   /* try to distribute timeouts somewhat randomly */
   ts.tv_nsec = ((unsigned long)self & 1023UL) * (1000000000UL / 1024UL);
 
   for (;;)
     {
-      ts.tv_sec  = time (0) + IDLE_TIMEOUT;
-
       X_LOCK (reqlock);
 
       for (;;)
@@ -813,6 +1029,7 @@ X_THREAD_PROC (eio_proc)
 
           ++idle;
 
+          ts.tv_sec = time (0) + IDLE_TIMEOUT;
           if (X_COND_TIMEDWAIT (reqwait, reqlock, ts) == ETIMEDOUT)
             {
               if (idle > max_idle)
@@ -827,7 +1044,6 @@ X_THREAD_PROC (eio_proc)
 
               /* we are allowed to idle, so do so without any timeout */
               X_COND_WAIT (reqwait, reqlock);
-              ts.tv_sec  = time (0) + IDLE_TIMEOUT;
             }
 
           --idle;
@@ -837,109 +1053,11 @@ X_THREAD_PROC (eio_proc)
 
       X_UNLOCK (reqlock);
      
-      errno = 0; /* strictly unnecessary */
+      if (req->type < 0)
+        goto quit;
 
       if (!EIO_CANCELLED (req))
-        switch (req->type)
-          {
-            case EIO_READ:      ALLOC (req->size);
-                                req->result = req->offs >= 0
-                                            ? pread     (req->int1, req->ptr2, req->size, req->offs)
-                                            : read      (req->int1, req->ptr2, req->size); break;
-            case EIO_WRITE:     req->result = req->offs >= 0
-                                            ? pwrite    (req->int1, req->ptr2, req->size, req->offs)
-                                            : write     (req->int1, req->ptr2, req->size); break;
-
-            case EIO_READAHEAD: req->result = readahead     (req->int1, req->offs, req->size); break;
-            case EIO_SENDFILE:  req->result = eio__sendfile (req->int1, req->int2, req->offs, req->size, self); break;
-
-            case EIO_STAT:      ALLOC (sizeof (EIO_STRUCT_STAT));
-                                req->result = stat      (req->ptr1, (EIO_STRUCT_STAT *)req->ptr2); break;
-            case EIO_LSTAT:     ALLOC (sizeof (EIO_STRUCT_STAT));
-                                req->result = lstat     (req->ptr1, (EIO_STRUCT_STAT *)req->ptr2); break;
-            case EIO_FSTAT:     ALLOC (sizeof (EIO_STRUCT_STAT));
-                                req->result = fstat     (req->int1, (EIO_STRUCT_STAT *)req->ptr2); break;
-
-            case EIO_CHOWN:     req->result = chown     (req->ptr1, req->int2, req->int3); break;
-            case EIO_FCHOWN:    req->result = fchown    (req->int1, req->int2, req->int3); break;
-            case EIO_CHMOD:     req->result = chmod     (req->ptr1, (mode_t)req->int2); break;
-            case EIO_FCHMOD:    req->result = fchmod    (req->int1, (mode_t)req->int2); break;
-            case EIO_TRUNCATE:  req->result = truncate  (req->ptr1, req->offs); break;
-            case EIO_FTRUNCATE: req->result = ftruncate (req->int1, req->offs); break;
-
-            case EIO_OPEN:      req->result = open      (req->ptr1, req->int1, (mode_t)req->int2); break;
-            case EIO_CLOSE:     req->result = close     (req->int1); break;
-            case EIO_DUP2:      req->result = dup2      (req->int1, req->int2); break;
-            case EIO_UNLINK:    req->result = unlink    (req->ptr1); break;
-            case EIO_RMDIR:     req->result = rmdir     (req->ptr1); break;
-            case EIO_MKDIR:     req->result = mkdir     (req->ptr1, (mode_t)req->int2); break;
-            case EIO_RENAME:    req->result = rename    (req->ptr1, req->ptr2); break;
-            case EIO_LINK:      req->result = link      (req->ptr1, req->ptr2); break;
-            case EIO_SYMLINK:   req->result = symlink   (req->ptr1, req->ptr2); break;
-            case EIO_MKNOD:     req->result = mknod     (req->ptr1, (mode_t)req->int2, (dev_t)req->offs); break;
-
-            case EIO_READLINK:  ALLOC (NAME_MAX);
-                                req->result = readlink  (req->ptr1, req->ptr2, NAME_MAX); break;
-
-            case EIO_SYNC:      req->result = 0; sync (); break;
-            case EIO_FSYNC:     req->result = fsync     (req->int1); break;
-            case EIO_FDATASYNC: req->result = fdatasync (req->int1); break;
-
-            case EIO_READDIR:   eio__scandir (req, self); break;
-
-            case EIO_BUSY:
-#ifdef _WIN32
-	      Sleep (req->nv1 * 1000.);
-#else
-              {
-                struct timeval tv;
-
-                tv.tv_sec  = req->nv1;
-                tv.tv_usec = (req->nv1 - tv.tv_sec) * 1000000.;
-
-                req->result = select (0, 0, 0, 0, &tv);
-              }
-#endif
-              break;
-
-            case EIO_UTIME:
-            case EIO_FUTIME:
-              {
-                struct timeval tv[2];
-                struct timeval *times;
-
-                if (req->nv1 != -1. || req->nv2 != -1.)
-                  {
-                    tv[0].tv_sec  = req->nv1;
-                    tv[0].tv_usec = (req->nv1 - tv[0].tv_sec) * 1000000.;
-                    tv[1].tv_sec  = req->nv2;
-                    tv[1].tv_usec = (req->nv2 - tv[1].tv_sec) * 1000000.;
-
-                    times = tv;
-                  }
-                else
-                  times = 0;
-
-
-                req->result = req->type == EIO_FUTIME
-                              ? futimes (req->int1, times)
-                              : utimes  (req->ptr1, times);
-              }
-
-            case EIO_GROUP:
-            case EIO_NOP:
-              req->result = 0;
-              break;
-
-            case EIO_QUIT:
-              goto quit;
-
-            default:
-              req->result = -1;
-              break;
-          }
-
-      req->errorno = errno;
+        ETP_EXECUTE (self, req);
 
       X_LOCK (reslock);
 
@@ -949,14 +1067,14 @@ X_THREAD_PROC (eio_proc)
         want_poll_cb ();
 
       self->req = 0;
-      worker_clear (self);
+      etp_worker_clear (self);
 
       X_UNLOCK (reslock);
     }
 
 quit:
   X_LOCK (wrklock);
-  worker_free (self);
+  etp_worker_free (self);
   X_UNLOCK (wrklock);
 
   return 0;
@@ -964,79 +1082,9 @@ quit:
 
 /*****************************************************************************/
 
-static void eio_atfork_prepare (void)
-{
-  X_LOCK (wrklock);
-  X_LOCK (reqlock);
-  X_LOCK (reslock);
-#if !HAVE_PREADWRITE
-  X_LOCK (preadwritelock);
-#endif
-#if !HAVE_READDIR_R
-  X_LOCK (readdirlock);
-#endif
-}
-
-static void eio_atfork_parent (void)
-{
-#if !HAVE_READDIR_R
-  X_UNLOCK (readdirlock);
-#endif
-#if !HAVE_PREADWRITE
-  X_UNLOCK (preadwritelock);
-#endif
-  X_UNLOCK (reslock);
-  X_UNLOCK (reqlock);
-  X_UNLOCK (wrklock);
-}
-
-static void eio_atfork_child (void)
-{
-  eio_req *prv;
-
-  while (prv = reqq_shift (&req_queue))
-    eio_destroy (prv);
-
-  while (prv = reqq_shift (&res_queue))
-    eio_destroy (prv);
-
-  while (wrk_first.next != &wrk_first)
-    {
-      worker *wrk = wrk_first.next;
-
-      if (wrk->req)
-        eio_destroy (wrk->req);
-
-      worker_clear (wrk);
-      worker_free (wrk);
-    }
-
-  started  = 0;
-  idle     = 0;
-  nreqs    = 0;
-  nready   = 0;
-  npending = 0;
-
-  eio_atfork_parent ();
-}
-
 int eio_init (void (*want_poll)(void), void (*done_poll)(void))
 {
-  want_poll_cb = want_poll;
-  done_poll_cb = done_poll;
-
-#ifdef _WIN32
-  X_MUTEX_CHECK (wrklock);
-  X_MUTEX_CHECK (reslock);
-  X_MUTEX_CHECK (reqlock);
-  X_MUTEX_CHECK (reqwait);
-  X_MUTEX_CHECK (preadwritelock);
-  X_MUTEX_CHECK (readdirlock);
-
-  X_COND_CHECK  (reqwait);
-#endif
-
-  X_THREAD_ATFORK (eio_atfork_prepare, eio_atfork_parent, eio_atfork_child);
+  etp_init (want_poll, done_poll);
 }
 
 static void eio_api_destroy (eio_req *req)
@@ -1051,9 +1099,10 @@ static void eio_api_destroy (eio_req *req)
   if (!req)                                                     \
     return 0;                                                   \
                                                                 \
-  req->type = rtype;                                            \
-  req->pri = EIO_DEFAULT_PRI + EIO_PRI_BIAS;                    \
-  req->finish = cb;						\
+  req->type    = rtype;                                         \
+  req->pri     = pri;						\
+  req->finish  = cb;						\
+  req->data    = data;						\
   req->destroy = eio_api_destroy;
 
 #define SEND eio_submit (req); return req
@@ -1067,159 +1116,268 @@ static void eio_api_destroy (eio_req *req)
       return 0;							\
     }
 
-eio_req *eio_nop (eio_cb cb)
+static void eio_execute (etp_worker *self, eio_req *req)
+{
+  errno = 0;
+
+  switch (req->type)
+    {
+      case EIO_READ:      ALLOC (req->size);
+                          req->result = req->offs >= 0
+                                      ? pread     (req->int1, req->ptr2, req->size, req->offs)
+                                      : read      (req->int1, req->ptr2, req->size); break;
+      case EIO_WRITE:     req->result = req->offs >= 0
+                                      ? pwrite    (req->int1, req->ptr2, req->size, req->offs)
+                                      : write     (req->int1, req->ptr2, req->size); break;
+
+      case EIO_READAHEAD: req->result = readahead     (req->int1, req->offs, req->size); break;
+      case EIO_SENDFILE:  req->result = eio__sendfile (req->int1, req->int2, req->offs, req->size, self); break;
+
+      case EIO_STAT:      ALLOC (sizeof (EIO_STRUCT_STAT));
+                          req->result = stat      (req->ptr1, (EIO_STRUCT_STAT *)req->ptr2); break;
+      case EIO_LSTAT:     ALLOC (sizeof (EIO_STRUCT_STAT));
+                          req->result = lstat     (req->ptr1, (EIO_STRUCT_STAT *)req->ptr2); break;
+      case EIO_FSTAT:     ALLOC (sizeof (EIO_STRUCT_STAT));
+                          req->result = fstat     (req->int1, (EIO_STRUCT_STAT *)req->ptr2); break;
+
+      case EIO_CHOWN:     req->result = chown     (req->ptr1, req->int2, req->int3); break;
+      case EIO_FCHOWN:    req->result = fchown    (req->int1, req->int2, req->int3); break;
+      case EIO_CHMOD:     req->result = chmod     (req->ptr1, (mode_t)req->int2); break;
+      case EIO_FCHMOD:    req->result = fchmod    (req->int1, (mode_t)req->int2); break;
+      case EIO_TRUNCATE:  req->result = truncate  (req->ptr1, req->offs); break;
+      case EIO_FTRUNCATE: req->result = ftruncate (req->int1, req->offs); break;
+
+      case EIO_OPEN:      req->result = open      (req->ptr1, req->int1, (mode_t)req->int2); break;
+      case EIO_CLOSE:     req->result = close     (req->int1); break;
+      case EIO_DUP2:      req->result = dup2      (req->int1, req->int2); break;
+      case EIO_UNLINK:    req->result = unlink    (req->ptr1); break;
+      case EIO_RMDIR:     req->result = rmdir     (req->ptr1); break;
+      case EIO_MKDIR:     req->result = mkdir     (req->ptr1, (mode_t)req->int2); break;
+      case EIO_RENAME:    req->result = rename    (req->ptr1, req->ptr2); break;
+      case EIO_LINK:      req->result = link      (req->ptr1, req->ptr2); break;
+      case EIO_SYMLINK:   req->result = symlink   (req->ptr1, req->ptr2); break;
+      case EIO_MKNOD:     req->result = mknod     (req->ptr1, (mode_t)req->int2, (dev_t)req->offs); break;
+
+      case EIO_READLINK:  ALLOC (NAME_MAX);
+                          req->result = readlink  (req->ptr1, req->ptr2, NAME_MAX); break;
+
+      case EIO_SYNC:      req->result = 0; sync (); break;
+      case EIO_FSYNC:     req->result = fsync     (req->int1); break;
+      case EIO_FDATASYNC: req->result = fdatasync (req->int1); break;
+
+      case EIO_READDIR:   eio__scandir (req, self); break;
+
+      case EIO_BUSY:
+#ifdef _WIN32
+	Sleep (req->nv1 * 1000.);
+#else
+        {
+          struct timeval tv;
+
+          tv.tv_sec  = req->nv1;
+          tv.tv_usec = (req->nv1 - tv.tv_sec) * 1000000.;
+
+          req->result = select (0, 0, 0, 0, &tv);
+        }
+#endif
+        break;
+
+      case EIO_UTIME:
+      case EIO_FUTIME:
+        {
+          struct timeval tv[2];
+          struct timeval *times;
+
+          if (req->nv1 != -1. || req->nv2 != -1.)
+            {
+              tv[0].tv_sec  = req->nv1;
+              tv[0].tv_usec = (req->nv1 - tv[0].tv_sec) * 1000000.;
+              tv[1].tv_sec  = req->nv2;
+              tv[1].tv_usec = (req->nv2 - tv[1].tv_sec) * 1000000.;
+
+              times = tv;
+            }
+          else
+            times = 0;
+
+
+          req->result = req->type == EIO_FUTIME
+                        ? futimes (req->int1, times)
+                        : utimes  (req->ptr1, times);
+        }
+
+      case EIO_GROUP:
+      case EIO_NOP:
+        req->result = 0;
+        break;
+
+      case EIO_CUSTOM:
+        req->feed (req);
+        break;
+
+      default:
+        req->result = -1;
+        break;
+    }
+
+  req->errorno = errno;
+}
+
+#ifndef EIO_NO_WRAPPERS
+
+eio_req *eio_nop (int pri, eio_cb cb, void *data)
 {
   REQ (EIO_NOP); SEND;
 }
 
-eio_req *eio_busy (double delay, eio_cb cb)
+eio_req *eio_busy (double delay, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_BUSY); req->nv1 = delay; SEND;
 }
 
-eio_req *eio_sync (eio_cb cb)
+eio_req *eio_sync (int pri, eio_cb cb, void *data)
 {
   REQ (EIO_SYNC); SEND;
 }
 
-eio_req *eio_fsync (int fd, eio_cb cb)
+eio_req *eio_fsync (int fd, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FSYNC); req->int1 = fd; SEND;
 }
 
-eio_req *eio_fdatasync (int fd, eio_cb cb)
+eio_req *eio_fdatasync (int fd, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FDATASYNC); req->int1 = fd; SEND;
 }
 
-eio_req *eio_close (int fd, eio_cb cb)
+eio_req *eio_close (int fd, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_CLOSE); req->int1 = fd; SEND;
 }
 
-eio_req *eio_readahead (int fd, off_t offset, size_t length, eio_cb cb)
+eio_req *eio_readahead (int fd, off_t offset, size_t length, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_READAHEAD); req->int1 = fd; req->offs = offset; req->size = length; SEND;
 }
 
-eio_req *eio_read (int fd, void *data, size_t length, off_t offset, eio_cb cb)
+eio_req *eio_read (int fd, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data)
 {
-  REQ (EIO_READ); req->int1 = fd; req->offs = offset; req->size = length; req->ptr2 = data; SEND;
+  REQ (EIO_READ); req->int1 = fd; req->offs = offset; req->size = length; req->ptr2 = buf; SEND;
 }
 
-eio_req *eio_write (int fd, void *data, size_t length, off_t offset, eio_cb cb)
+eio_req *eio_write (int fd, void *buf, size_t length, off_t offset, int pri, eio_cb cb, void *data)
 {
-  REQ (EIO_WRITE); req->int1 = fd; req->offs = offset; req->size = length; req->ptr2 = data; SEND;
+  REQ (EIO_WRITE); req->int1 = fd; req->offs = offset; req->size = length; req->ptr2 = buf; SEND;
 }
 
-eio_req *eio_fstat (int fd, eio_cb cb)
+eio_req *eio_fstat (int fd, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FSTAT); req->int1 = fd; SEND;
 }
 
-eio_req *eio_futime (int fd, double atime, double mtime, eio_cb cb)
+eio_req *eio_futime (int fd, double atime, double mtime, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FUTIME); req->int1 = fd; req->nv1 = atime; req->nv2 = mtime; SEND;
 }
 
-eio_req *eio_ftruncate (int fd, off_t offset, eio_cb cb)
+eio_req *eio_ftruncate (int fd, off_t offset, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FTRUNCATE); req->int1 = fd; req->offs = offset; SEND;
 }
 
-eio_req *eio_fchmod (int fd, mode_t mode, eio_cb cb)
+eio_req *eio_fchmod (int fd, mode_t mode, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FCHMOD); req->int1 = fd; req->int2 = (long)mode; SEND;
 }
 
-eio_req *eio_fchown (int fd, uid_t uid, gid_t gid, eio_cb cb)
+eio_req *eio_fchown (int fd, uid_t uid, gid_t gid, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FCHOWN); req->int1 = fd; req->int2 = (long)uid; req->int3 = (long)gid; SEND;
 }
 
-eio_req *eio_dup2 (int fd, int fd2, eio_cb cb)
+eio_req *eio_dup2 (int fd, int fd2, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_DUP2); req->int1 = fd; req->int2 = fd2; SEND;
 }
 
-eio_req *eio_sendfile (int out_fd, int in_fd, off_t in_offset, size_t length, eio_cb cb)
+eio_req *eio_sendfile (int out_fd, int in_fd, off_t in_offset, size_t length, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_SENDFILE); req->int1 = out_fd; req->int2 = in_fd; req->offs = in_offset; req->size = length; SEND;
 }
 
-eio_req *eio_open (const char *path, int flags, mode_t mode, eio_cb cb)
+eio_req *eio_open (const char *path, int flags, mode_t mode, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_OPEN); PATH; req->int1 = flags; req->int2 = (long)mode; SEND;
 }
 
-eio_req *eio_utime (const char *path, double atime, double mtime, eio_cb cb)
+eio_req *eio_utime (const char *path, double atime, double mtime, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_UTIME); PATH; req->nv1 = atime; req->nv2 = mtime; SEND;
 }
 
-eio_req *eio_truncate (const char *path, off_t offset, eio_cb cb)
+eio_req *eio_truncate (const char *path, off_t offset, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_TRUNCATE); PATH; req->offs = offset; SEND;
 }
 
-eio_req *eio_chown (const char *path, uid_t uid, gid_t gid, eio_cb cb)
+eio_req *eio_chown (const char *path, uid_t uid, gid_t gid, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_CHOWN); PATH; req->int2 = (long)uid; req->int3 = (long)gid; SEND;
 }
 
-eio_req *eio_chmod (const char *path, mode_t mode, eio_cb cb)
+eio_req *eio_chmod (const char *path, mode_t mode, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_CHMOD); PATH; req->int2 = (long)mode; SEND;
 }
 
-eio_req *eio_mkdir (const char *path, mode_t mode, eio_cb cb)
+eio_req *eio_mkdir (const char *path, mode_t mode, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_MKDIR); PATH; req->int2 = (long)mode; SEND;
 }
 
 static eio_req *
-eio__1path (int type, const char *path, eio_cb cb)
+eio__1path (int type, const char *path, int pri, eio_cb cb, void *data)
 {
   REQ (type); PATH; SEND;
 }
 
-eio_req *eio_readlink (const char *path, eio_cb cb)
+eio_req *eio_readlink (const char *path, int pri, eio_cb cb, void *data)
 {
-  return eio__1path (EIO_READLINK, path, cb);
+  return eio__1path (EIO_READLINK, path, pri, cb, data);
 }
 
-eio_req *eio_stat (const char *path, eio_cb cb)
+eio_req *eio_stat (const char *path, int pri, eio_cb cb, void *data)
 {
-  return eio__1path (EIO_STAT, path, cb);
+  return eio__1path (EIO_STAT, path, pri, cb, data);
 }
 
-eio_req *eio_lstat (const char *path, eio_cb cb)
+eio_req *eio_lstat (const char *path, int pri, eio_cb cb, void *data)
 {
-  return eio__1path (EIO_LSTAT, path, cb);
+  return eio__1path (EIO_LSTAT, path, pri, cb, data);
 }
 
-eio_req *eio_unlink (const char *path, eio_cb cb)
+eio_req *eio_unlink (const char *path, int pri, eio_cb cb, void *data)
 {
-  return eio__1path (EIO_UNLINK, path, cb);
+  return eio__1path (EIO_UNLINK, path, pri, cb, data);
 }
 
-eio_req *eio_rmdir (const char *path, eio_cb cb)
+eio_req *eio_rmdir (const char *path, int pri, eio_cb cb, void *data)
 {
-  return eio__1path (EIO_RMDIR, path, cb);
+  return eio__1path (EIO_RMDIR, path, pri, cb, data);
 }
 
-eio_req *eio_readdir (const char *path, eio_cb cb)
+eio_req *eio_readdir (const char *path, int pri, eio_cb cb, void *data)
 {
-  return eio__1path (EIO_READDIR, path, cb);
+  return eio__1path (EIO_READDIR, path, pri, cb, data);
 }
 
-eio_req *eio_mknod (const char *path, mode_t mode, dev_t dev, eio_cb cb)
+eio_req *eio_mknod (const char *path, mode_t mode, dev_t dev, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_MKNOD); PATH; req->int2 = (long)mode; req->int2 = (long)dev; SEND;
 }
 
 static eio_req *
-eio__2path (int type, const char *path, const char *new_path, eio_cb cb)
+eio__2path (int type, const char *path, const char *new_path, int pri, eio_cb cb, void *data)
 {
   REQ (type); PATH;
 
@@ -1234,19 +1392,33 @@ eio__2path (int type, const char *path, const char *new_path, eio_cb cb)
   SEND;
 }
 
-eio_req *eio_link (const char *path, const char *new_path, eio_cb cb)
+eio_req *eio_link (const char *path, const char *new_path, int pri, eio_cb cb, void *data)
 {
-  return eio__2path (EIO_LINK, path, new_path, cb);
+  return eio__2path (EIO_LINK, path, new_path, pri, cb, data);
 }
 
-eio_req *eio_symlink (const char *path, const char *new_path, eio_cb cb)
+eio_req *eio_symlink (const char *path, const char *new_path, int pri, eio_cb cb, void *data)
 {
-  return eio__2path (EIO_SYMLINK, path, new_path, cb);
+  return eio__2path (EIO_SYMLINK, path, new_path, pri, cb, data);
 }
 
-eio_req *eio_rename (const char *path, const char *new_path, eio_cb cb)
+eio_req *eio_rename (const char *path, const char *new_path, int pri, eio_cb cb, void *data)
 {
-  return eio__2path (EIO_RENAME, path, new_path, cb);
+  return eio__2path (EIO_RENAME, path, new_path, pri, cb, data);
+}
+
+eio_req *eio_custom (eio_cb execute, int pri, eio_cb cb, void *data)
+{
+  REQ (EIO_CUSTOM); req->feed = execute; SEND;
+}
+
+#endif
+
+eio_req *eio_grp (eio_cb cb, void *data)
+{
+  const int pri = EIO_PRI_MAX;
+
+  REQ (EIO_GROUP); SEND;
 }
 
 #undef REQ
@@ -1292,7 +1464,7 @@ void eio_grp_add (eio_req *grp, eio_req *req)
 
 ssize_t eio_sendfile_sync (int ofd, int ifd, off_t offset, size_t count)
 {
-  worker wrk;
+  etp_worker wrk;
 
   wrk.dbuf = 0;
 
